#include <avr/io.h>
#include <avr/interrupt.h>


// для теста
byte ii;

#define ZERO_CROSS       2  //номер пина сигнала детектора нуля (дискретный вход)
#define TRIAC_GATE_PULSE 22   //номер пина сигнала управления триаком (дискретный выход)
#define VOLTAGE_ADC      3   //номер аналогового пина измерения мгновенного напряжения 
                             //при переназначении необходимо изменить значение ADMUX в разделе setup

#define widthTRIAC_GATE_PULSE 1   //длительность импульса управления триаком в тиках таймера3
                                  //при делителе таймера3 = 256 (16МГц/256=62500Гц = 16мкс/тик)
                                  //длительность импульса = 4*16мкс=64мкс
                                  
#define samples_frequencyVOLTAGE_MEASUREMENT 10    // период выборки значений напряжения в тиках Таймера1 для 62-х выборок
                                                   //в течение полуволны переменного напряжения
                                                   // Измерения запускаются по прерыванию переполнения Таймера 1
                                                   // Частота полуволны переменнного напряжения 50Гц составляет 100Гц 
                                                   // В течение полуволны Таймер1 выдает 62500Гц/100Гц=625 тиков (10мс)
                                                   // Для обеспечения 62 выборок в течение полуволны выборка должна запускаться
                                                   // c периодом 625/60=10 тик (160 мксек)
                                                   

byte statusVOLTAGE_MEASUREMENT = 0; //0-не активное, 1-запрос на начало выборки, 2-в процессе выборки, 3-выборка закончена, 4-обработка результата, 5-результат обработан, можно следующий
                                    //измерения вызываются ХХ раз в секунду, измеряются значения в пределе полупериода и накапливаются в массив
                                    //частота выборки в пределе полуволны определятется значением компаратора счетчика OCR1B
                                    //старт измерений происходит по прерыванию перехода напряжения через 0
#define max_array_indexVOLTAGE_MEASUREMENT 70//максимальное кол-во элементов массива хранения выборки знаений измеренного напряжения 
byte array_indexVOLTAGE_MEASUREMENT; //индекс массива хранения выборки значений измеренного напряжения
unsigned int arrayVOLTAGE_MEASUREMENT[max_array_indexVOLTAGE_MEASUREMENT];//массив хранения выборки значений измеренного напряжения в течение полупериода волны

void setup(){
  ii=0;
  Serial.begin(9600);//инициализация com1 для отладки
  // установка пинов
  pinMode(ZERO_CROSS, INPUT);     //определение направляения сигнала
  digitalWrite(ZERO_CROSS, HIGH); //включение подтягивающего резистора
  pinMode(TRIAC_GATE_PULSE, OUTPUT);      //определение направления сигнала

  ////////////////////////////
  //// установка Таймера3 ////
  ////////////////////////////  
  // Таймер3 используется для организации фазового управления напряжением (мощностью)
  // Последоватлеьность работы
  // * таймер установлен но запрещен
  // * обнаружен сигнал пересечения нуля на 2
  // * таймер начинает считать с нуля
  // * компаратор установлен на значение задержки включения триака
  // * счетчик таймера достигает значения компаратора
  // * обработчик прерывания компаратора включает триак (подачет сигнал на вывод GATE)
  // * счетчик таймера устанавливает длиительность импульса включения триака (разрядность счетчика минус длительность импульсов)
  // * счетчик таймера достигает переполнения
  // * обработчик прерывания по перевыполнению снимает управляющий сигнал с триака
  // * триак закрывается сам при следующем переходе через ноль (особенность работы триаков)
  // Частота процессора 16MHz. 
  // Использование делителя 256 обеспечивает длительность тика 16 мкс 16МГц/256=62500Гц = 16мкс/тик)
  // Частота полуволны переменнного напряжения 50Гц составляет 100Гц 
  // В течение полуволны таймер выдает 625 тиков (10мс)
  TIMSK3 &= (0<<OCIE3B); //1-разрешение прерывания по компаратору В
  TIMSK3 |= (1<<OCIE3A); //1-разрешение прерывания по компаратору А
  TIMSK3 |= (1<<TOIE3); //1-разрешение прерывания по переполнению    
  TCCR3A = 0x00;    //установка регистров контроля таймера (timer control register) для
  TCCR3B = 0x00;    //нормальной работы, таймер запрещен
  OCR3A = 460;      //инициализация значения компаратора
  
  ////////////////////////////
  //// установка Таймера1 ////
  ////////////////////////////  
  // Таймер1 задает частоту опроса АЦП
  // Частота процессора 16MHz. 
  // Использование делителя 256 обеспечивает длительность тика 16 мкс 16МГц/256=62500Гц = 16мкс/тик)
  // Частота полуволны переменнного напряжения 50Гц составляет 100Гц 
  // В течение полуволны таймер выдает 62500Гц/100Гц=625 тиков (10мс)
  // Для обеспечения 60 выборок в течение полуволны значние компаратора должно быть 625/60=10 тик (160 мксек)
  TIMSK1 &= (0<<OCIE1B); //1-разрешение прерывания по компаратору В
  TIMSK1 &= (0<<OCIE1A); //1-разрешение прерывания по компаратору А
  TIMSK1 |= (1<<TOIE1); //1-разрешение прерывания по переполнению  
  TCCR1A = 0x00;    //установка регистров контроля таймера (timer control register) для
  TCCR1B = 0x00;    //нормальной работы, таймер запрещен
  
  /////////////////////////// 
  //// Инициализация АЦП ////
  ///////////////////////////
//  DIDR0=0xFF; //запретить дискретные входы на аналоговых пинах 7-0
//  DIDR2=0xFF; //запретить дискретные входы на аналоговых пинах 15-8
 
  //инициализация мультиплексора
  //бит 7-6 выбор опорного напряжения 2,56В
  //бит 5   установка право ориентированного порядка следования старшего и младшего байт АЦП
  //бит 4-0 выбор аналогового пина 3
  ADMUX = 0b11000011;
  
  //инициализация регистра ADCSRA – ADC Control and Status Register A
  //бит 7 разшершить АЦП   
  //бит 6 запустить преобразование
  //бит 5 разрешить преобразование по сигналу триггера
  //бит 4 флаг прерывания
  //бит 3 разшерить прерывания АЦП
  //бит 2-0 выбор делителя частоты (64) 
  ADCSRA = 0b11101110;
  
  //Установка регистра ADCSRB – ADC Control and Status Register B 
  // бит 6 Analog Multiplecsor Comparator Enable
  // бит 3 выбор аналогового пина 3
  // бит 2-0 выбор триггера для АЦП преобразования - переполнение Таймера 1
  ADCSRB = 0b01000110;
  
  // инициализация прерывания при переходе нуля напряжения
  // прерывание по переднему фронту сигнала, привязанного к пину 2
  attachInterrupt(0,zeroCrossingInterrupt, RISING);    

  //разрешение прерываний
  sei();
}  

//Обработчик прерывания перехода нуля
void zeroCrossingInterrupt(){ //обнаружен переход нуля   
  TCCR3B=0x04; //запуск таймера 3 с делителем 256 (16МГц/256=62500Гц = 16мкс/тик)
  TCNT3 = 0;   //обнуление значения таймера 3
  
  if (statusVOLTAGE_MEASUREMENT == 1)  //если получен запрос на начало выборки измерения напряжения - начинаем выборку
  {
    TCCR1B=0x04; //запуск таймера 1 с делителем 256 (16МГц/256=62500Гц = 16мкс/тик)
    TCNT1 = 65536 - samples_frequencyVOLTAGE_MEASUREMENT; //устанавливаем значение таймера 1
    array_indexVOLTAGE_MEASUREMENT = 0; //обнуляем индекс выборки массива данных
    statusVOLTAGE_MEASUREMENT =2; //выборка значений напряжения полуволны в процессе
  } else
    {
      if (statusVOLTAGE_MEASUREMENT == 2)  //если идет процесс измерения напряжения и вызывается повторное прерывание по переходу нуля
      {                                     //заканчиваем выборку
        TCCR1B=0x00; //останавливаем таймер 1
        statusVOLTAGE_MEASUREMENT =3; //выборка значений напряжения полуволны закончена
      }
    }
}

ISR(TIMER3_COMPA_vect){ //обработчик прерывания по достижению таймером 3 значения компаратора А
  PORTA |= (1<<PORTA0);   //подача управляющего сигнала на вход GATE тиака (PIN22)  
  TCNT3 = 65536 - widthTRIAC_GATE_PULSE; //установка значения таймера на длительность импульса управляющего сигнала триака
                                       //управляющий сигнал будет отключен по прерыванию переполнения таймера
}

//Обработчик прерывания по переполнению таймера 3
ISR(TIMER3_OVF_vect){ 
  PORTA &= (0<<PORTA0);   //снятие управляющего сигнала с входа GATE тиака (PIN22)
  TCCR3B = 0x00;          //запрещение таймера для избежания ложных срабатываний
}

ISR(TIMER1_OVF_vect){ //обработчик прерывания по достижению таймером 3 значения компаратора B
  TCNT1 = 65536 - samples_frequencyVOLTAGE_MEASUREMENT;   //установка значения таймера 1 для следующего вызова прерывания по переполнению
}

//Обработчик прерывания окончания АЦП преобразования
//данный обработчик считывает значение текущего напряжения с аналогового пина 3 в массив выборок
ISR(ADC_vect){
  arrayVOLTAGE_MEASUREMENT[array_indexVOLTAGE_MEASUREMENT] = ADCL;//сохранение младшего байта АЦП
  arrayVOLTAGE_MEASUREMENT[array_indexVOLTAGE_MEASUREMENT] += ADCH << 8;//сохранение старшего байта АЦП
  array_indexVOLTAGE_MEASUREMENT ++; //увеличение индекса на 1
}


void loop(){ // sample code to exercise the circuit
//Для отладки, выводит значения массива выборки напряжения в ком порт  при получении любого значения из порта.
  if (statusVOLTAGE_MEASUREMENT == 0)statusVOLTAGE_MEASUREMENT = 1;
  if (statusVOLTAGE_MEASUREMENT == 3) //если 
  {
    if(ii<62)
    { 
       Serial.println(arrayVOLTAGE_MEASUREMENT[ii]);  
       ii++;
    }   else 
          {
            statusVOLTAGE_MEASUREMENT = 4; 
            ii=0;
         }
  }
  if (Serial.available()>0) //для если в ком-порт записан любой байт (например, через программу гипертерминал), запускаем процесс измерения напряжения.
  {
    statusVOLTAGE_MEASUREMENT = 0; 
    ii=0; 
    Serial.read();
  }  
}
